#!/usr/bin/env python3
import copy
import json
import sys
from typing import List, Dict
import time

import pkg_resources
import singer
from pymongo import MongoClient
from singer import metadata, metrics, utils
from c8connector import (
    C8Connector, ConfigProperty, Sample, Schema,
    ConfigAttributeType, SchemaAttribute, ValidationException)

from macrometa_source_mongo.sync_strategies import log_based
from macrometa_source_mongo.sync_strategies import common
from macrometa_source_mongo.sync_strategies import full_table
from macrometa_source_mongo.sync_strategies.sample_data import fetch_samples
from macrometa_source_mongo.connection import create_certficate_files, delete_certficate_files, \
    get_connection_string, get_user_databases
from macrometa_source_mongo.exceptions import InvalidReplicationMethodException, NoReadPrivilegeException
from macrometa_source_mongo.helper import filter_streams_by_replication_method, get_attribute_type, \
    get_streams_to_sync, produce_collection_schema, streams_list_to_dict, \
    write_schema_message

LOGGER = singer.get_logger('macrometa_source_mongo')

REQUIRED_CONFIG_KEYS = [
    'host',
    'user',
    'password',
    'auth_database',
    'database',
    'source_collection'
]

REQUIRED_CONFIG_KEYS_NON_SRV = REQUIRED_CONFIG_KEYS + ['port']

LOG_BASED_METHOD = 'LOG_BASED'
FULL_TABLE_METHOD = 'FULL_TABLE'


class MongoSourceConnector(C8Connector):
    """MongoSourceConnector's C8Connector impl."""

    def name(self) -> str:
        """Returns the name of the connector."""
        return "MongoDB"

    def package_name(self) -> str:
        """Returns the package name of the connector (i.e. PyPi package name)."""
        return "macrometa-source-mongo"

    def version(self) -> str:
        """Returns the version of the connector."""
        return pkg_resources.get_distribution('macrometa_source_mongo').version

    def type(self) -> str:
        """Returns the type of the connector."""
        return "source"

    def description(self) -> str:
        """Returns the description of the connector."""
        return "Source data from a MongoDB collection."

    def validate(self, integration: dict) -> None:
        """Validate given configurations against the connector.
        If invalid, throw an exception with the cause.
        """
        config = self.get_config(integration)
        try:
            config = create_certficate_files(config)
            client = self.get_client(config)
            do_discover(client, config)
        except Exception as e:
            delete_certficate_files(config)
            raise e
        delete_certficate_files(config)

    def samples(self, integration: dict) -> list[Sample]:
        """Fetch sample data using the provided configurations."""
        config = self.get_config(integration)
        try:
            config = create_certficate_files(config)
            client = self.get_client(config)
            streams = do_discover(client, config)
            results = []
            for stream in streams:
                schema, data = fetch_samples(client, config, stream)
                # Appending _ to keys inorder for preserving values of reserved keys in source data
                schema['__id'] = schema.pop('_id')
                data = data[:10]
                for record in data:
                    record['__id'] = record.pop('_id')
                results.append(Sample(
                    schema=Schema(stream["table_name"],
                                  [SchemaAttribute(k, get_attribute_type(v)) for k, v in schema.items()]),
                    data=data
                ))
        except Exception as e:
            delete_certficate_files(config)
            raise e
        delete_certficate_files(config)
        return results

    def schemas(self, integration: dict) -> list[Schema]:
        """Get supported schemas using the given configurations."""
        config = self.get_config(integration)
        try:
            config = create_certficate_files(config)
            client = self.get_client(config)
            streams = do_discover(client, config)
            results = []
            for stream in streams:
                schema, _ = fetch_samples(client, config, stream)
                # Appending _ to keys inorder for preserving values of reserved keys in source data
                schema['__id'] = schema.pop('_id')
                results.append(Schema(stream["table_name"],
                                      [SchemaAttribute(k, get_attribute_type(v)) for k, v in schema.items()]))
        except Exception as e:
            delete_certficate_files(config)
            raise e
        delete_certficate_files(config)
        return results

    def logo(self) -> str:
        """Returns the logo image for the connector."""
        return ""

    def reserved_keys(self) -> list[str]:
        """List of reserved keys for the connector."""
        return []

    def config(self) -> list[ConfigProperty]:
        """Get configuration parameters for the connector."""
        return [
            ConfigProperty('host', 'Host', ConfigAttributeType.STRING, True, False,
                           description='MongoDB host.',
                           placeholder_value='mongodb_host'),
            ConfigProperty('port', 'Port', ConfigAttributeType.INT, False, False,
                           description='MongoDB port.',
                           default_value='27017'),
            ConfigProperty('user', 'Username', ConfigAttributeType.STRING, True, False,
                           description='MongoDB user.',
                           placeholder_value='mongo'),
            ConfigProperty('password', 'Password', ConfigAttributeType.PASSWORD, True, False,
                           description='MongoDB password.',
                           placeholder_value='password'),
            ConfigProperty('auth_database', 'Auth Database', ConfigAttributeType.STRING, True, False,
                           description='MongoDB authentication database.',
                           default_value='admin'),
            ConfigProperty('database', 'Database', ConfigAttributeType.STRING, True, True,
                           description='MongoDB database name.',
                           placeholder_value='mongo'),
            ConfigProperty('source_collection', 'Source Collection', ConfigAttributeType.STRING, True, True,
                           description="Source collection name.", placeholder_value="my_collection"),
            ConfigProperty('replication_method', 'Replication Method', ConfigAttributeType.STRING, False, False,
                           description='Choose from LOG_BASED, FULL_TABLE.',
                           default_value='FULL_TABLE'),
            ConfigProperty('srv', 'Enable SRV', ConfigAttributeType.BOOLEAN, False, False,
                           description='Uses a `mongodb+srv` protocol to connect. Disables the usage of `port` '
                                       'argument if set to `True`.',
                           default_value="false"),
            ConfigProperty('replica_set', 'Replica Set', ConfigAttributeType.STRING, False, False,
                           description='Name of replica set.',
                           placeholder_value='replica'),
            ConfigProperty('ssl', 'Use SSL', ConfigAttributeType.BOOLEAN, False, False,
                           description='Can be set to true to connect using SSL.',
                           default_value="false"),
            ConfigProperty('verify_mode', 'Verify Mode', ConfigAttributeType.BOOLEAN, False, False,
                           description='Default SSL verify mode.',
                           default_value="true"),
            ConfigProperty('direct_connection', 'Direct Connection', ConfigAttributeType.BOOLEAN, False, False,
                           description='Specifies whether to connect directly to the specified MongoDB host as a '
                                       'standalone or connect to the entire replica set of which the given MongoDB host is a part.',
                           default_value="false"),
            ConfigProperty('tls_ca_file', 'SSL/TLS CA Certificate', ConfigAttributeType.FILE, False, False,
                           description='Specific CA certificate in PEM string format. This is most often the case '
                                       'when using `self-signed` server certificate.',
                           placeholder_value="my_ca_certificate"),
            ConfigProperty('tls_certificate_key_file', 'SSL/TLS Client Certificate', ConfigAttributeType.FILE, False, False,
                           description='Specific client certificate in PEM string format. If the private key for the client '
                                       'certificate is stored in a separate file, it should be concatenated with the certificate file.',
                           placeholder_value="my_client_certificate"),
            ConfigProperty('tls_certificate_key_file_password', 'SSL/TLS Client Key Password', ConfigAttributeType.PASSWORD, False, False,
                           description='If the private key contained in the certificate keyfile is encrypted, users can provide a '
                                       'password or passphrase to decrypt the encrypted private keys.',
                           placeholder_value="my_client_key_password"),
        ]

    def capabilities(self) -> list[str]:
        """Return the capabilities[1] of the connector.
        [1] https://docs.meltano.com/contribute/plugins#how-to-test-a-tap
        """
        return ['catalog', 'discover', 'state']

    @staticmethod
    def get_config(integration: dict) -> dict:
        try:
            return {
                # Required config keys
                'host': integration['host'],
                'user': integration['user'],
                'password': integration['password'],
                'database': integration['database'],
                'auth_database': integration['auth_database'],
                'source_collection': integration['source_collection'],
                # Optional config keys
                'replication_method': integration.get('replication_method', FULL_TABLE_METHOD),
                'srv': integration.get('srv', False),
                'port': integration.get('port'),
                'replica_set': integration.get('replica_set'),
                'ssl': integration.get('ssl', False),
                'verify_mode': integration.get('verify_mode', True),
                'direct_connection': integration.get('direct_connection', False),
                'tls_ca_file': integration.get('tls_ca_file'),
                'tls_certificate_key_file': integration.get('tls_certificate_key_file'),
                'tls_certificate_key_file_password': integration.get('tls_certificate_key_file_password'),
            }
        except KeyError as e:
            raise ValidationException(f'Integration property `{e}` not found.') from e

    @staticmethod
    def get_client(config: Dict) -> MongoClient:
        connection_string = get_connection_string(config)
        client = MongoClient(connection_string)
        LOGGER.info('Connected to MongoDB host: %s, version: %s',
                    config['host'],
                    client.server_info().get('version', 'unknown'))
        return client


def do_discover(client: MongoClient, config: Dict):
    """
    Executes discovery mode, scanning the MongoDB cluster and converting
    all collections in the specified database into streams. The result is
    output as JSON to stdout.
    Args:
        client (MongoClient): A MongoDB Client instance.
        config (Dict): A dictionary containing database configuration parameters.
    """
    try:
        if config['replication_method'] not in [LOG_BASED_METHOD, FULL_TABLE_METHOD]:
            raise Exception('Invalid replication method provided. It should be either FULL_TABLE or LOG_BASED.')

        if config['database'] not in get_user_databases(client, config):
            raise NoReadPrivilegeException(config['user'], config['database'])

        database = client[config['database']]
        collection_name = config['source_collection']
        # Check if collection exists
        if collection_name not in database.list_collection_names():
            raise Exception(f'Collection "{collection_name}" not found in the database {config["database"]}.')

        collection = database[collection_name]
        is_view = collection.options().get('viewOn') is not None

        if is_view:
            raise KeyError('The connector does not support views.')

        LOGGER.info("Retrieving collection information for database '%s', collection '%s'", database.name, collection_name)
        streams = [produce_collection_schema(collection)]
        json.dump({'streams': streams}, sys.stdout, indent=2)

        return streams
    except Exception as e:
        raise ValidationException(e)


def clear_state_on_replication_change(stream: Dict, state: Dict, replication_method: str) -> Dict:
    """
    Resets the state for a given stream if the replication method has changed.
    Args:
        stream (Dict): A dictionary representing a stream.
        state (Dict): The current state of the replication process.
        replication_method (str): The desired replication method.

    Returns:
        Dict: The updated state.
    """
    md_map = metadata.to_map(stream['metadata'])
    tap_stream_id = stream['tap_stream_id']

    # Change in replication method detected
    current_replication_method = md_map.get('replication-method', replication_method)
    last_replication_method = singer.get_bookmark(state, tap_stream_id, 'last_replication_method')
    if last_replication_method is not None and (current_replication_method != last_replication_method):
        log_msg = (f"Detected change in replication method for stream {tap_stream_id}. Changing from"
                   f" {last_replication_method} to {current_replication_method}. Re-replicating entire collection.")
        LOGGER.info(log_msg)
        state = singer.reset_stream(state, tap_stream_id)

    state = singer.write_bookmark(state, tap_stream_id, 'last_replication_method', current_replication_method)

    return state


def sync_traditional_stream(client: MongoClient, stream: Dict, state: Dict, replication_method: str):
    """
    Syncs the stream using a traditional replication method i.e. FULL_TABLE.
    Args:
        client (MongoClient): An instance of the MongoClient.
        stream (Dict): A dictionary representing the stream to be synced.
        state (Dict): The current state of the replication process.
        replication_method (str): The desired replication method.

    Raises:
        InvalidReplicationMethodException: If the replication method is not either FULL_TABLE or LOG_BASED.
    """
    tap_stream_id = stream['tap_stream_id']

    common.COUNTS[tap_stream_id] = 0
    common.TIMES[tap_stream_id] = 0
    common.SCHEMA_COUNT[tap_stream_id] = 0
    common.SCHEMA_TIMES[tap_stream_id] = 0

    md_map = metadata.to_map(stream['metadata'])
    replication_method = md_map.get('replication-method', replication_method)

    if replication_method not in {LOG_BASED_METHOD, FULL_TABLE_METHOD}:
        raise InvalidReplicationMethodException(replication_method,
                                                'replication method needs to be either FULL_TABLE'
                                                ' or LOG_BASED')

    database_name = metadata.get(md_map, (), 'database-name')

    # Emit a state message to indicate that we've started this stream
    state = clear_state_on_replication_change(stream, state, replication_method)
    state = singer.set_currently_syncing(state, stream['tap_stream_id'])
    singer.write_message(singer.StateMessage(value=copy.deepcopy(state)))

    if replication_method == LOG_BASED_METHOD:
        resume_token = singer.get_bookmark(state,
                                          stream['tap_stream_id'],
                                          log_based.BOOKMARK_KEY)
        start_after = resume_token.get('_data') if resume_token else None
        cursor = client[database_name].watch(
                        [{'$match': {
                            '$or': [
                                {'operationType': 'insert'}, {'operationType': 'update'}, {'operationType': 'delete'}
                            ],
                            '$and': [
                                # watch collections of selected streams
                                {'ns.coll': {'$in': [stream['table_name']]}}
                            ]
                        }}],
                        max_await_time_ms=500,
                        start_after=start_after
                    )
        cursor.try_next()
        # Get resume token from '_data' to resume LOG_BASED
        resume_token = {
            '_data': cursor.resume_token['_data']
        }
        cursor.close()
        state = singer.write_bookmark(state, stream['tap_stream_id'], 'token', resume_token)

    LOGGER.info("Starting full table sync")
    write_schema_message(stream)
    common.SCHEMA_COUNT[tap_stream_id] += 1

    with metrics.job_timer('sync_table') as timer:
        timer.tags['database'] = database_name
        timer.tags['table'] = stream['table_name']

        collection = client[database_name][stream["table_name"]]
        full_table.sync_database(collection, stream, state)

    LOGGER.info("Sleeping")
    time.sleep(100)
    LOGGER.info("Full table sync completed")
    state = singer.set_currently_syncing(state, None)

    singer.write_message(singer.StateMessage(value=copy.deepcopy(state)))


def sync_traditional_streams(client: MongoClient, traditional_streams: List[Dict], state: Dict,
                             replication_method: str):
    """
    This function syncs a list of traditional streams that use FULL_TABLE replication method, one stream at a time.
    Args:
        client (MongoClient): A MongoClient instance used to connect to the MongoDB database.
        traditional_streams (List[Dict]): A list of dictionaries, where each dictionary represents a traditional stream to be synced.
        state (Dict): A dictionary representing the current state of the sync.
        replication_method (str): A string representing the replication method to use for syncing the streams.
    """
    for stream in traditional_streams:
        sync_traditional_stream(client, stream, state, replication_method)


def sync_log_based_streams(client: MongoClient,
                           log_based_streams: List[Dict],
                           database_name: str,
                           state: Dict,
                           replication_method: str
                           ):
    """
    Sync log_based streams all at once by listening on the database-level change streams events.
    Args:
        client: MongoDB client instance
        log_based_streams:  list of streams to sync
        database_name: name of the database to sync from
        state: state dictionary
        replication_method: replication method
    """
    if not log_based_streams:
        return

    streams = streams_list_to_dict(log_based_streams)

    for tap_stream_id, stream in streams.items():
        common.COUNTS[tap_stream_id] = 0
        common.TIMES[tap_stream_id] = 0
        common.SCHEMA_COUNT[tap_stream_id] = 0
        common.SCHEMA_TIMES[tap_stream_id] = 0

        state = clear_state_on_replication_change(stream, state, replication_method)
        singer.write_message(singer.StateMessage(value=copy.deepcopy(state)))

        write_schema_message(stream)
        common.SCHEMA_COUNT[tap_stream_id] += 1

    with metrics.job_timer('sync_table') as timer:
        timer.tags['database'] = database_name

        log_based.sync_database(client[database_name], streams, state)

    state = singer.set_currently_syncing(state, None)
    singer.write_message(singer.StateMessage(value=copy.deepcopy(state)))


def do_sync(client: MongoClient, catalog: Dict, config: Dict, replication_method: str, state: Dict):
    """
    Syncs all the selected streams in the catalog
    Args:
        client: MongoDb client instance
        catalog: dictionary with all the streams details
        config: config dictionary
        replication_method: replication method
        state: state
    """

    all_streams = catalog['streams']
    streams_to_sync = get_streams_to_sync(all_streams, state)

    log_based_streams, traditional_streams = filter_streams_by_replication_method(streams_to_sync,
                                                                                  replication_method)
    LOGGER.debug('Starting sync of traditional streams ...')
    sync_traditional_streams(client, traditional_streams, state, replication_method)
    LOGGER.debug('Sync of traditional streams done')

    LOGGER.debug('Starting sync of log based streams ...')
    sync_log_based_streams(client,
                           log_based_streams,
                           config['database'],
                           state,
                           replication_method
                           )
    LOGGER.debug('Sync of log based streams done')

    LOGGER.info(common.get_sync_summary(catalog, replication_method))


def main_impl():
    """
    Main function
    """
    args = utils.parse_args(REQUIRED_CONFIG_KEYS)
    config = args.config
    srv = config.get('srv', False)

    if not srv:
        args = utils.parse_args(REQUIRED_CONFIG_KEYS_NON_SRV)
        config = args.config
    config['replication_method']: args.config.get('replication_method', FULL_TABLE_METHOD)
    try:
        config = create_certficate_files(config)
        connection_string = get_connection_string(config)
        client = MongoClient(connection_string)

        LOGGER.info('Connected to MongoDB host: %s, version: %s',
                    config['host'],
                    client.server_info().get('version', 'unknown'))

        if args.discover:
            do_discover(client, config)
        elif args.catalog:
            state = args.state or {}
            do_sync(client, args.catalog.to_dict(), config, args.config.get('replication_method', FULL_TABLE_METHOD), state)
    except Exception as e:
        delete_certficate_files(config)
        raise e
    delete_certficate_files(config)


def main():
    """
    Main
    """
    try:
        main_impl()
    except Exception as exc:
        LOGGER.exception(exc)
        raise exc



#!/usr/bin/env python3
import base64
import datetime
import time
import uuid
import bson
import singer
import pytz
import tzlocal

from typing import Dict, Any, Optional
from bson import objectid, timestamp, datetime as bson_datetime
from singer import utils, metadata
from terminaltables import AsciiTable

from macrometa_source_mongo.exceptions import InvalidDateTimeException, SyncException, UnsupportedKeyTypeException

SDC_DELETED_AT = "_sdc_deleted_at"
COUNTS = {}
TIMES = {}
SCHEMA_COUNT = {}
SCHEMA_TIMES = {}


def should_sync_column(md_map, field_name):
    field_metadata = md_map.get(('properties', field_name), {})
    return singer.should_sync_field(field_metadata.get('inclusion'),
                                    field_metadata.get('selected'),
                                    True)


def calculate_destination_stream_name(stream: Dict) -> str:
    """
    Constructs the appropriate stream name for use in Singer messages.

    Args:
        stream: A dictionary containing stream information.
    
    Returns:
        A string representing the stream name.
    """
    s_md = metadata.to_map(stream['metadata'])
    return f"{s_md.get((), {}).get('database-name')}-{stream['stream']}"


def get_stream_version(tap_stream_id: str, state: Dict) -> int:
    """
    Retrieves the stream version from the given state dictionary or generates a new one if not present.
    Args:
        tap_stream_id: The stream ID for which the version is needed.
        state: The state dictionary from which to extract the version, if it exists.

    Returns:
       The version as an integer.
    """
    stream_version = singer.get_bookmark(state, tap_stream_id, 'version')

    if stream_version is None:
        stream_version = int(time.time() * 1000)

    return stream_version


def class_to_string(key_value: Any, key_type: str) -> str:
    """
    Transforms various data types into their string representations.
    Supported types include: datetime, bson Timestamp, bytes, int, Int64, float, ObjectId, str, and UUID.
    Args:
        key_value: The value to be converted into a string.
        key_type: The data type of the value.

    Returns:
        The string representation of the given value.

    Raises:
        UnsupportedKeyTypeException: If the specified key_type is not supported.
    """
    if key_type == 'datetime':
        if key_value.tzinfo is None:
            timezone = tzlocal.get_localzone()
            local_datetime = timezone.localize(key_value)
            utc_datetime = local_datetime.astimezone(pytz.UTC)
        else:
            utc_datetime = key_value.astimezone(pytz.UTC)

        return utils.strftime(utc_datetime)

    if key_type == 'Timestamp':
        return f'{key_value.time}.{key_value.inc}'

    if key_type == 'bytes':
        return base64.b64encode(key_value).decode('utf-8')

    if key_type in {'int', 'Int64', 'float', 'ObjectId', 'str', 'UUID'}:
        return str(key_value)

    raise UnsupportedKeyTypeException(f"The key type {key_value} is not supported")


def string_to_class(str_value: str, type_value: str) -> Any:
    """
    Converts the specified string value to the given type, if the type is supported.
    
    Supported types include: UUID, datetime, int, Int64, float, ObjectId, Timestamp, bytes, and str.
    
    Args:
        str_value (str): The string value to convert.
        type_value (str): The type of the value.
        
    Returns:
        Any: The converted string value.
    
    Raises:
        UnsupportedKeyTypeException: If the key is not supported.
    """
    conversion = {
        'UUID': uuid.UUID,
        'datetime': singer.utils.strptime_with_tz,
        'int': int,
        'Int64': bson.int64.Int64,
        'float': str,
        'ObjectId': objectid.ObjectId,
        'Timestamp': lambda val: (lambda split_value=val.split('.'):
                                  bson.timestamp.Timestamp(int(split_value[0]), int(split_value[1])))(),
        'bytes': lambda val: base64.b64decode(val.encode()),
        'str': str,
    }

    if type_value in conversion:
        return conversion[type_value](str_value)

    raise UnsupportedKeyTypeException(f"The key type {type_value} is not supported")


def safe_transform_datetime(value: datetime.datetime, path):
    """
    Transforms a datetime object safely from local time zone to UTC if possible.

    Args:
        value (datetime.datetime): The datetime object to be transformed.
        path (Any): The path to the value.

    Returns:
        str: The UTC datetime as a string.
    """
    timezone = tzlocal.get_localzone()
    try:
        local_datetime = timezone.localize(value)
        utc_datetime = local_datetime.astimezone(pytz.UTC)
    except Exception as ex:
        if str(ex) == "year is out of range" and value.year == 0:
            # formatting it as a string and passing it along
            return f"{value.year:04d}-{value.month:02d}-{value.day:02d}T{value.hour:02d}:{value.minute:02d}:" \
                   f"{value.second:02d}.{value.microsecond:06d}Z"
        raise InvalidDateTimeException(f"Datetime Invalid at [{'.'.join(map(str, path))}]: {value}") from ex
    return utils.strftime(utc_datetime)


def transform_value(value: Any, path) -> Any:
    """
    Transforms a value to a format suitable for JSON.

    Args:
        value (Any): The value to be transformed.
        path (Any): The path to the value.

    Returns:
        Any: The transformed value.
    """
    conversion = {
        list: lambda val, pat: list(map(lambda v: transform_value(v[1], pat + [v[0]]), enumerate(val))),
        dict: lambda val, pat: {k: transform_value(v, pat + [k]) for k, v in val.items()},
        uuid.UUID: lambda val, _: class_to_string(val, 'UUID'),
        objectid.ObjectId: lambda val, _: class_to_string(val, 'ObjectId'),
        bson_datetime.datetime: safe_transform_datetime,
        timestamp.Timestamp: lambda val, _: utils.strftime(val.as_datetime()),
        bson.int64.Int64: lambda val, _: class_to_string(val, 'Int64'),
        bytes: lambda val, _: class_to_string(val, 'bytes'),
        datetime.datetime: lambda val, _: class_to_string(val, 'datetime'),
        bson.decimal128.Decimal128: lambda val, _: val.to_decimal(),
        bson.regex.Regex: lambda val, _: dict(pattern=val.pattern, flags=val.flags),
        bson.code.Code: lambda val, _: dict(value=str(val), scope=str(val.scope)) if val.scope else str(val),
        bson.dbref.DBRef: lambda val, _: dict(id=str(val.id), collection=val.collection, database=val.database),
    }

    if isinstance(value, tuple(conversion.keys())):
        return conversion[type(value)](value, path)

    return value


def row_to_singer_record(stream: Dict,
                         row: Dict,
                         time_extracted: datetime.datetime,
                         time_deleted: Optional[datetime.datetime],
                         version: Optional[int] = None,
                         ) -> singer.RecordMessage:
    """
    Converts a row to a Singer RecordMessage instance.

    Args:
        stream (Dict): A dictionary containing details for the stream.
        row (Dict): A dictionary containing the row to be converted.
        time_extracted (datetime.datetime): The time when the row was extracted.
        time_deleted (Optional[datetime.datetime]): The time when the row was deleted, if applicable.
        version (Optional[int]): The version of the stream.
        
    Returns:
        singer.RecordMessage: A Singer RecordMessage instance.
    """

    if version is None:
        version = int(time.time() * 1000)

    try:
        row_to_persist = {k: transform_value(v, [k]) for k, v in row.items()
                          if not isinstance(v, (bson.min_key.MinKey, bson.max_key.MaxKey))}
    except InvalidDateTimeException as ex:
        raise SyncException(
            f"Encountered an error while syncing collection: {stream['tap_stream_id']},"
            " record ID: {row['_id']} - {ex}") from ex

    row_to_persist = {
        '_id': str(row_to_persist['_id']),
        'document': row_to_persist,
        SDC_DELETED_AT: utils.strftime(time_deleted) if time_deleted else None
    }

    return singer.RecordMessage(
        stream=calculate_destination_stream_name(stream),
        record=row_to_persist,
        version=version,
        time_extracted=time_extracted)


def get_sync_summary(catalog, replication_method: str)->str:
    """
    Constructs a summary of the synchronization for all streams.
    Args:
    catalog (dict): A dictionary containing details for all the streams.
    replication_method (str): The replication method being used.

    Returns: 
        str: A string representation of the summary table.
    """
    headers = [['database',
                'collection',
                'replication method',
                'total records',
                'write speed',
                'total time',
                'schemas written',
                'schema build duration',
                'percent building schemas']]

    rows = []
    for stream_id, stream_count in COUNTS.items():
        stream = [x for x in catalog['streams'] if x['tap_stream_id'] == stream_id][0]
        collection_name = stream.get("table_name")
        md_map = metadata.to_map(stream['metadata'])
        db_name = metadata.get(md_map, (), 'database-name')
        replication_method = md_map.get('replication-method', replication_method)

        stream_time = TIMES[stream_id]
        schemas_written = SCHEMA_COUNT[stream_id]
        schema_duration = SCHEMA_TIMES[stream_id]

        if stream_time == 0:
            stream_time = 0.000001

        rows.append(
            [
                db_name,
                collection_name,
                replication_method,
                f'{stream_count} records',
                f'{stream_count / float(stream_time):.1f} records/second',
                f'{stream_time:.5f} seconds',
                f'{schemas_written} schemas',
                f'{schema_duration:.5f} seconds',
                f'{100 * schema_duration / float(stream_time):.2f}%'
            ]
        )

    data = headers + rows
    table = AsciiTable(data, title='Sync Summary')

    return '\n\n' + table.table



import bson
import pymongo

from singer import metadata
from macrometa_source_mongo.sync_strategies import common

from macrometa_source_mongo.exceptions import InvalidDateTimeException, SyncException

def fetch_table(client, conn_info, stream):
    """
    Fetches data from a MongoDB collection for a specified stream.

    Args:
        client (pymongo.MongoClient): A MongoClient instance connected to the MongoDB server.
        conn_info (dict): A dictionary containing connection information, such as host, port, and database name.
        stream (dict): A dictionary containing information about the stream, such as table name and stream ID.

    Returns:
        Tuple containing the schema for the stream and data fetched from the MongoDB collection.
    """
    database = client[conn_info['dbname']]
    collection = database[stream["table_name"]]
    data = []
    rec = {}
    cursor = collection.find({}, sort=[("_id", pymongo.ASCENDING)]).limit(10)

    for row in cursor:
        try:
            rec = {k: common.transform_value(v, [k]) for k, v in row.items()
                              if not isinstance(v, (bson.min_key.MinKey, bson.max_key.MaxKey))}
        except InvalidDateTimeException as ex:
            raise SyncException(
                f"Error syncing collection {stream['tap_stream_id']}, object ID {row['_id']} - {ex}") from ex

        rec = {
            '_id': str(rec['_id']),
            'document': rec
        }
        data.append(rec)

    schema = {"_id": "string", "document": "object"}
    return schema, data


def fetch_samples(client, conn_config, stream):
    """
    Fetches samples of data from a MongoDB collection for a specified stream.

    Args:
        client (pymongo.MongoClient): A MongoClient instance connected to the MongoDB server.
        conn_config (dict): A dictionary containing connection configuration information.
        stream (dict): A dictionary containing information about the stream, such as table name and stream ID.

    Returns:
        Tuple containing the schema for the stream and sample data fetched from the MongoDB collection.
    """
    md_map = metadata.to_map(stream['metadata'])
    conn_config['dbname'] = md_map.get(()).get('database-name')
    return fetch_table(client, conn_config, stream)
